var ts = require("typescript");
var typehelper = require("./typehelper");
var utils = require("./utils");
//#region Private Method
function addTypeToTag(comments, tag, node, checker, type) {
    var nodeType = node.type;
    if (!nodeType && !type) {
        tag.type = "*";
        return;
    }
    var typeName = typehelper.getNodeTypeName(node, nodeType, checker, type);
    typeName.anonymous.forEach(function (subNode) {
        var comm = comments.addLinked(), tagName = utils.getAnonymousName(node);
        if (subNode.parameters) {
            var signature = checker.getSignatureFromDeclaration(subNode);
            comm.addTag("callback", tagName);
            addParent(comm, node, checker);
            addFromSignature(comm, subNode, signature, checker);
        }
        else {
            var subType = checker.getTypeAtLocation(subNode), calls = subType.getCallSignatures(), ctors = subType.getConstructSignatures();
            if (calls.length || ctors.length) {
                comm.addTag("interface", tagName);
                addParentModule(comm, subNode, checker);
                addCallSignatures(comm, subNode, "ctor", ctors, checker);
                addCallSignatures(comm, subNode, "call", calls, checker);
            }
            else {
                comm.addTag("typedef", tagName).type = "Object";
                addParent(comm, node, checker);
            }
        }
    });
    if (node.kind === 123 /* Parameter */ && node.dotDotDotToken) {
        typeName.name = utils.formatType("..." + typeName.name);
    }
    tag.type = typeName.name;
}
function addParam(comments, param, checker) {
    var paramDeclaration = param.valueDeclaration, paramName = utils.formatParam(param.name, paramDeclaration), paramDesc = utils.getComments(param), tag = comments.getParamTag(param.name, paramDeclaration);
    tag.name = paramName;
    addTypeToTag(comments, tag, paramDeclaration, checker);
    if (paramDesc && paramDesc !== tag.desc) {
        tag.desc = paramDesc;
    }
}
//#endregion
//#region Generic Methods
function addName(comments, tagName, node, checker) {
    var tag = comments.getOrAddTag(tagName), nodeType = node.type, initializer = node.initializer;
    tag.name = utils.getNodeName(node);
    if (checker) {
        if (nodeType) {
            addTypeToTag(comments, tag, node, checker);
        }
        if (initializer && initializer.kind === 6 /* FirstLiteralToken */) {
            if (tagName === "member") {
                comments.getOrAddTag("default").name = initializer.getText();
            }
            else {
                tag.name += "=" + initializer.getText();
            }
        }
        if (initializer && !nodeType) {
            addTypeToTag(comments, tag, null, checker, initializer.contextualType);
        }
    }
}
exports.addName = addName;
function addGeneric(comments, node) {
    var params = node.typeParameters;
    if (params && params.length > 0) {
        var tag = comments.getOrAddTag("generic");
        tag.name = params.map(function (t) { return utils.getNodeName(t); }).join(", ");
    }
}
exports.addGeneric = addGeneric;
function addParent(comments, node, checker) {
    if (!node.parent) {
        return;
    }
    var parentName = utils.getParentName(node, checker);
    if (parentName) {
        var tag = comments.getOrAddTag("memberof");
        tag.name = parentName;
    }
}
exports.addParent = addParent;
function addParentModule(comments, node, checker) {
    if (!node.parent) {
        return;
    }
    var parentName = utils.getParentModuleName(node, checker);
    if (parentName) {
        var tag = comments.getOrAddTag("memberof");
        tag.name = parentName;
    }
}
exports.addParentModule = addParentModule;
function addModifier(comments, node) {
    var isInterface = node.parent && node.parent.kind === 186 /* InterfaceDeclaration */;
    if (!isInterface && node.modifiers) {
        if ((node.modifiers.flags & 16 /* Public */) === 16 /* Public */) {
            comments.getOrAddTag("public");
        }
        else if ((node.modifiers.flags & 32 /* Private */) === 32 /* Private */) {
            comments.getOrAddTag("private");
        }
        else if ((node.modifiers.flags & 64 /* Protected */) === 64 /* Protected */) {
            comments.getOrAddTag("protected");
        }
        if ((node.modifiers.flags & 128 /* Static */) === 128 /* Static */) {
            comments.getOrAddTag("static");
        }
    }
    if (node.parent && (node.parent.kind === 185 /* ClassDeclaration */ || node.parent.kind === 186 /* InterfaceDeclaration */)) {
        comments.getOrAddTag("instance");
    }
}
exports.addModifier = addModifier;
function addDefault(comments, node) {
    var defaultValue = utils.getDefault(node);
    if (defaultValue) {
        comments.getOrAddTag("default").name = defaultValue;
    }
}
exports.addDefault = addDefault;
function merge(comments, node, checker) {
    var typeName = utils.getParentName(node, checker), nodeName = utils.getNodeName(node) || typeName, prevComments, cache = comments.list.cache.merges;
    typeName = typeName ? typeName + "." + nodeName : nodeName;
    if (!cache) {
        cache = comments.list.cache.merges = Object.create(null);
    }
    if (typeof cache[typeName] === "undefined") {
        cache[typeName] = comments;
    }
    else {
        prevComments = cache[typeName];
        comments.mergeTo(prevComments);
        comments.clean();
    }
}
exports.merge = merge;
function addVariation(comments, node, checker) {
    var typeName = utils.getParentName(node, checker), nodeName = utils.getNodeName(node) || typeName, prevComments, cache = comments.list.cache.variations;
    typeName = typeName ? typeName + "." + nodeName : nodeName;
    if (!cache) {
        cache = comments.list.cache.variations = Object.create(null);
    }
    if (typeof cache[typeName] === "undefined") {
        cache[typeName] = 1;
    }
    else {
        cache[typeName]++;
        comments.getOrAddTag("variation").name = cache[typeName];
    }
}
exports.addVariation = addVariation;
//#endregion
//#region Specialized Methods
function addFromSignature(comments, node, signature, checker) {
    var params = signature.getParameters(), returns = signature.getReturnType(), comms = utils.getComments(signature), tag;
    if (comms && comms !== comments.desc) {
        comments.desc = comms;
    }
    if (params) {
        params.forEach(function (param) { return addParam(comments, param, checker); });
    }
    if (returns) {
        tag = comments.getOrAddTag("returns");
        addTypeToTag(comments, tag, node, checker);
    }
}
exports.addFromSignature = addFromSignature;
function addImplementations(comments, node, checker) {
    if (!node.heritageClauses) {
        return;
    }
    node.heritageClauses.forEach(function (clause) {
        var tagName = clause.token === 77 /* ExtendsKeyword */ ? "augments" : "implements";
        clause.types.forEach(function (clauseType) {
            var typeResult = typehelper.getNodeTypeName(node, clauseType, checker);
            comments.getOrAddTag(tagName, typeResult.name);
        });
    });
}
exports.addImplementations = addImplementations;
function addCallbacks(comments, node, signatures, checker) {
    signatures.forEach(function (signature) {
        var comm = comments.clone(), sigNode = signature.getDeclaration();
        addName(comm, "callback", node);
        addGeneric(comm, node);
        addFromSignature(comm, sigNode, signature, checker);
        addParent(comm, node, checker);
        addVariation(comm, node, checker);
    });
}
exports.addCallbacks = addCallbacks;
function addCallSignatures(comments, node, tagPrefix, signatures, checker) {
    signatures.forEach(function (signature) {
        var comm = comments.addLinked(), childNode = signature.getDeclaration();
        comm.getOrAddTag(tagPrefix + "signature");
        addGeneric(comm, childNode);
        addFromSignature(comm, childNode, signature, checker);
        addParent(comm, childNode, checker);
        addVariation(comm, childNode, checker);
    });
}
exports.addCallSignatures = addCallSignatures;
function addEnumMembers(comments, node, checker) {
    node.members.forEach(function (member) {
        //var comm = comments.addLinked();
        var name = utils.getNodeName(member), tag = comments.getOrAddTag("property", name);
        if (member.initializer && member.initializer.kind === 6 /* FirstLiteralToken */) {
            tag.name += "=" + member.initializer.getText();
        }
        //addName(comm, "member", member);
        //addParent(comm, member, checker);
        //addDefault(comm, member);
    });
}
exports.addEnumMembers = addEnumMembers;
function addTypeDefProperty(comments, tagName, node, checker) {
    var name = utils.getNodeName(node), tag = comments.getOrAddTag(tagName, name), nodeType = node.type;
    if (nodeType) {
        var type = checker.getTypeAtLocation(nodeType);
        addTypeToTag(comments, tag, node, checker);
    }
}
exports.addTypeDefProperty = addTypeDefProperty;
//#endregion
