var getCommentsRegexp = /\/\*\*\s*((?!\s*\*\/)[\s\S]+?)\s*\*\//m, getCommentsDescRegexp = /\s*\*?\s*((?:(?!\s*\*\s*@\w+)[\s\S])+)/m, getCommentsTagRegexp = /[^\{]@([^\s]+)\s*(?:\{(.+?)\})?\s*([^\s]+)?((?:(?!\s*\*\s*@\w+)[\s\S])+)?/gm;
var Tag = (function () {
    function Tag(tag, name, type, desc) {
        this.tag = tag;
        this.name = name;
        this.type = type;
        this.desc = desc;
    }
    Tag.fromText = function (text) {
        var tags = [], tagMatch;
        while (tagMatch = getCommentsTagRegexp.exec(text)) {
            tags.push(new Tag(tagMatch[1], tagMatch[3], tagMatch[2], tagMatch[4]));
        }
        return tags;
    };
    Tag.prototype.toString = function () {
        var result = " * @" + this.tag;
        if (this.type) {
            result += " {" + this.type + "}";
        }
        if (this.name) {
            result += " " + this.name;
        }
        if (this.desc) {
            result += " " + this.desc;
        }
        return result;
    };
    return Tag;
})();
exports.Tag = Tag;
var Comments = (function () {
    function Comments() {
        this.raw = null;
        this.content = null;
        this.desc = null;
        this.tags = [];
    }
    Comments.fromNode = function (node) {
        var text = node.getFullText();
        return Comments.fromText(text);
    };
    Comments.fromText = function (text) {
        var match = text.match(getCommentsRegexp), result = new Comments(), descMatch, tagsMatch;
        if (!match) {
            return result;
        }
        result.raw = match[0];
        result.content = match[1];
        result.tags = [];
        if (descMatch = result.content.match(getCommentsDescRegexp)) {
            result.desc = descMatch[1];
        }
        result.tags = Tag.fromText(result.content);
        return result;
    };
    Comments.prototype.isEmpty = function () {
        return (!this.desc && this.tags.length === 0);
    };
    Comments.prototype.toString = function () {
        if (this.isEmpty()) {
            return "";
        }
        var result = "\n/**";
        if (this.desc) {
            result += "\n * " + this.desc;
        }
        result += "\n" + this.tags.map(function (t) { return t.toString(); }).join("\n");
        return result + "\n */";
    };
    Comments.prototype.hasTag = function (tag, name) {
        return !!this.getTag(tag, name);
    };
    Comments.prototype.getTag = function (tag, name) {
        if (this.tags.length === 0) {
            return null;
        }
        var i = 0, len = this.tags.length, current;
        for (; i < len; i++) {
            current = this.tags[i];
            if (current.tag === tag && (!name || current.name === name)) {
                return current;
            }
        }
        return null;
    };
    Comments.prototype.addTag = function (tag, name, type, desc) {
        var tmpTag = new Tag(tag, name, type, desc);
        this.tags.push(tmpTag);
        return tmpTag;
    };
    Comments.prototype.getOrAddTag = function (tag, name) {
        var tmp = this.getTag(tag, name);
        if (!tmp) {
            tmp = this.addTag(tag, name);
        }
        return tmp;
    };
    Comments.prototype.getParamTag = function (name, node) {
        if (this.tags.length === 0) {
            return this.addTag("param", name);
        }
        var tag = this.getTag("param", name), repeat = !!node && !!node.dotDotDotToken, optional = !!node && (!!node.questionToken || !!node.initializer);
        if (!tag && repeat && optional) {
            tag = this.getTag("param", "[..." + name + "]");
        }
        if (!tag && repeat) {
            tag = this.getTag("param", "..." + name);
        }
        if (!tag && optional) {
            tag = this.getTag("param", "[" + name + "]");
        }
        if (!tag) {
            tag = this.addTag("param", name);
        }
        return tag;
    };
    Comments.prototype.clone = function () {
        var comment = new Comments();
        comment.desc = this.desc;
        comment.tags = this.tags.slice(0);
        if (this.list) {
            this.list.add(comment);
        }
        return comment;
    };
    Comments.prototype.mergeTo = function (other) {
        if (this.desc) {
            other.desc = other.desc ? other.desc + "\n *\n * " : "";
            other.desc += this.desc;
        }
        this.tags.forEach(function (tag) {
            var otherTag = other.getOrAddTag(tag.tag, tag.name);
            if (tag.desc && tag.desc !== otherTag.desc) {
                otherTag.desc = tag.desc;
            }
            if (tag.type && tag.desc !== otherTag.type) {
                otherTag.type = tag.type;
            }
        });
    };
    Comments.prototype.clean = function () {
        this.desc = null;
        this.tags = [];
    };
    Comments.prototype.addLinked = function () {
        if (!this.list) {
            throw new Error("This comment is not linked to any list");
        }
        return this.list.add();
    };
    return Comments;
})();
exports.Comments = Comments;
var CommentsList = (function () {
    function CommentsList() {
        this.cache = {};
        this.list = [];
    }
    CommentsList.prototype.add = function (comment) {
        if (this.list.indexOf(comment) !== -1) {
            return;
        }
        if (!comment) {
            comment = new Comments();
        }
        comment.list = this;
        this.list.push(comment);
        return comment;
    };
    CommentsList.prototype.addFromNode = function (node) {
        var comment = Comments.fromNode(node);
        return this.add(comment);
    };
    CommentsList.prototype.get = function (kind, name, parentName) {
        var result;
        this.list.some(function (comment) {
            if (comment.hasTag(kind, name) && (!parentName || comment.hasTag("memberof", parentName))) {
                result = comment;
                return true;
            }
        });
        return result;
    };
    CommentsList.prototype.clear = function () {
        this.list = [];
    };
    CommentsList.prototype.toString = function () {
        return this.list.filter(function (item) { return !item.isEmpty(); }).join("\n\n").replace(/\n{2,}/g, "\n\n");
    };
    return CommentsList;
})();
exports.CommentsList = CommentsList;
